/*
 * asm.c
 * -----
 * N7 assembler compiler. Compiles n7a text files to binary files that can be
 * executed by the vm (renv). No effort has been put into this code. It's an 
 * ugly mess of cut-and-paste but gets the job done. I'll fix it ... later.
 *
 * By: Marcus 2021
 */
 
#include "stdlib.h"
#include "stdio.h"
#include "ctype.h"
#include "string.h"

#include "asm.h"
#include "bytecodes.h"
#include "hash_table.h"
#include "renv.h"

/* Token types. */
#define ASM_REG 1
#define ASM_VAR 2
#define ASM_NUM 3
#define ASM_STR 4
#define ASM_LBL 5
#define ASM_CMD 6
#define ASM_ERR 7
#define ASM_EOF 8
#define ASM_IGNORE 9

/* Base commands, translated to different instructions based on parameters. */
typedef enum {
    CMD_NOP = 1,
    CMD_END,
    CMD_MDUMP,
    CMD_RDUMP,
    CMD_SDUMP,
    CMD_MADD,
    CMD_MLOAD,
    CMD_MSET,
    CMD_MGET,
    CMD_MPUSH,      
    CMD_MPOP,
    CMD_MSWAP,
    CMD_MOVE,
    CMD_JMP,
    CMD_EVAL,
    CMD_JMPT,
    CMD_JMPF,
    CMD_PUSH,
    CMD_POP,
    CMD_SWAP,
    CMD_SPOP,
    CMD_OR,
    CMD_AND,
    CMD_NOT,
    CMD_EQL,
    CMD_LESS,
    CMD_GRE,
    CMD_LEQL,
    CMD_GEQL,
    CMD_NEQL,
    CMD_ADD,
    CMD_SUB,
    CMD_MUL,
    CMD_DIV,
    CMD_MOD,
    CMD_NEG,
    CMD_CTBL,
    CMD_TOSTR,
    CMD_TONUM,
    CMD_TOINT,
    CMD_ABS,
    CMD_MDEL,
    CMD_GC,
    CMD_LGC,    /* Lock GC, prevent it from collecting. */
    CMD_ULGC,   /* Unlock GC. */
    CMD_CPY,    /* Deep copy variable. */
    CMD_ECMP,
    CMD_RTE,
    CMD_CLR,
    CMD_CALL,
    CMD_RET,
    CMD_LOCAL,
    CMD_MCLR,
    CMD_LOADPM,
    CMD_SIZE,
    CMD_LEN,
    CMD_ILOAD,
    CMD_IHAS,
    CMD_IVAL,
    CMD_IKEY,
    CMD_IPUSH,
    CMD_IPOP,
    CMD_ISTEP,
    CMD_IDEL,
    CMD_COS,
    CMD_SIN,
    CMD_TAN,
    CMD_ACOS,
    CMD_ASIN,
    CMD_ATAN,
    CMD_ATAN2,
    CMD_SQR,
    CMD_POW,
    CMD_FLOOR,
    CMD_CEIL,
    CMD_ROUND,
    CMD_RAD,
    CMD_DEG,
    CMD_MIN,
    CMD_MAX,

    CMD_TYPE,
    CMD_MLOADS,
    CMD_LPTBL,
    CMD_SYS     /* System function. */
} Command;

#define INSTRUCTIONS_GROW_SIZE 1024
#define LINE_NUMBERS_GROW_SIZE 512
#define FILENAMES_GROW_SIZE    64

static FILE *sFile;

/* Table containing commands. The struct should be replaced with something that
   lists related bytecode instructions and their parameter types. This could be
   used to cleanup the compiler loop (a copy and paste mess).
     I'll fix that ...later. */
typedef struct {
    Command command;
} CommandEntry;
static HashTable *sCommands;
static void CreateCommands();
static void DeleteCommands();
static const char *GetCommandString(Command command);

/* Token data. */
static int sNext;
unsigned char sCommand;
static double sNumber;
static int sRegister;
static int sVariable;
static int sString;
static int sLabel;
static char sLabelName[64];
static char sError[256];

/* Instructions. */
Instruction *sInstructions;
int sInstructionsCapacity;
int sInstructionCount;
static void AddInstruction(Instruction instruction);

/* String constants. */
typedef struct {
    int index;
} StringEntry;
HashTable *sStrings;
char **sStringTable;
int sStringCount;
int AddString(const char *str);
static void StringEntryToTable(char *skey, int ikey, void *data, void *userData);
static void DeleteStringEntry(void *data);

/* Labels are gathered and translated to instruction indexes for jumps. */
typedef struct {
    int index;
    int instruction;
} LabelEntry;
HashTable *sLabels;
int sLabelId;
static int AddLabel(const char *lbl);
static void LinkLabel(char *skey, int ikey, void *data, void *userData);
static void DeleteLabelEntry(void *data);

/* Meta data, for generating more valuable runtime errors. */
static LineNumberMetadata *sLineNumbers;
static int sLineNumbersCapacity;
static int sLineNumberCount;
static void AddLineNumberMetadata(int lineNumber);

static FilenameMetadata *sFilenames;
static int sFilenamesCapacity;
static int sFilenameCount;
static void AddFilenameMetadata(const char *filename);

static void CleanUp();
static int GetNext();

/*
 * ASM_Compile
 * -----------
 * Compile assembler file srcFilename to binary file dstFilename.
 */
int ASM_Compile(const char *srcFilename, const char *dstFilename) {    
    sFile = fopen(srcFilename, "r");
    FILE *out;
    
    strcpy(sError, "");
    
    if (sFile == 0) {
        sprintf(sError, "Error: Could not open file %s", srcFilename);
        return ASM_FAILURE;
    }
    
    CreateCommands();
      
    sStrings = HT_Create(1);
    sStringCount = 0;
    sStringTable = 0;
    
    sLabels = HT_Create(1);
    sLabelId = 1;
    
    sInstructions = 0;
    sInstructionsCapacity = 0;
    sInstructionCount = 0;

    sLineNumbers = 0;
    sLineNumbersCapacity = 0;
    sLineNumberCount = 0;
    
    sFilenames = 0;
    sFilenamesCapacity = 0;
    sFilenameCount = 0;
    
    sNext = GetNext();
    while (!(sNext == ASM_ERR || sNext == ASM_EOF)) {
        if (sNext == ASM_CMD) {
            /* Create a struct with parameter types for each vm instruction
               instead, and get rid of this copy-and-paste-mess. */
            Instruction instruction;
            if (sCommand == CMD_NOP) {
                instruction.cmd = BC_NOP;
            }
            else if (sCommand == CMD_END) {
                instruction.cmd = BC_END;
            }
            else if (sCommand == CMD_MDUMP) {
                instruction.cmd = BC_MDUMP;
            }
            else if (sCommand == CMD_RDUMP) {
                instruction.cmd = BC_RDUMP;
            }
            else if (sCommand == CMD_SDUMP) {
                instruction.cmd = BC_SDUMP;
            }
            else if (sCommand == CMD_MADD) {
                int lparam = GetNext();
                /*int rparam;*/
                if (lparam == ASM_STR) {
                    instruction.cmd = BC_MADD_S;
                    instruction.lparam.i = sString;
                }
                else if (lparam == ASM_VAR) {
                    instruction.cmd = BC_MADD_S;
                    instruction.lparam.i = sVariable;
                }
                else if (lparam == ASM_NUM) {
                    instruction.cmd = BC_MADD_N;
                    instruction.lparam.i = (int)sNumber;
                }
                else if (lparam == ASM_REG) {
                    instruction.cmd = BC_MADD_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_MADD);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_MSET) {
                int lparam = GetNext();
                if (lparam == ASM_STR) {
                    instruction.cmd = BC_MSET_S;
                    instruction.lparam.i = sString;
                }
                else if (lparam == ASM_NUM) {
                    instruction.cmd = BC_MSET_N;
                    instruction.lparam.d = sNumber;
                }
                else if (lparam == ASM_LBL) {
                    instruction.cmd = BC_MSET_L;
                    instruction.lparam.i = -sLabel;
                }
                else if (lparam == ASM_REG) {
                    instruction.cmd = BC_MSET_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_MSET);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_MGET) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_MGET_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_MGET);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_MLOAD) {
                int lparam = GetNext();
                if (lparam == ASM_STR) {
                    instruction.cmd = BC_MLOAD_S;
                    instruction.lparam.i = sString;
                }
                else if (lparam == ASM_VAR) {
                    instruction.cmd = BC_MLOAD_S;
                    instruction.lparam.i = sVariable;
                }
                else if (lparam == ASM_NUM) {
                    instruction.cmd = BC_MLOAD_N;
                    instruction.lparam.i = (int)sNumber;
                }
                else if (lparam == ASM_REG) {
                    instruction.cmd = BC_MLOAD_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_MLOAD);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_MLOADS) {
                instruction.cmd = BC_MLOADS;
            }
            else if (sCommand == CMD_LOADPM) {
                instruction.cmd = BC_MLOAD;
            }
            else if (sCommand == CMD_MPUSH) {
                instruction.cmd = BC_MPUSH;
            }
            else if (sCommand == CMD_MPOP) {
                instruction.cmd = BC_MPOP;
            }
            else if (sCommand == CMD_MSWAP) {
                instruction.cmd = BC_MSWAP;
            }
            else if (sCommand == CMD_MOVE) {
                int lparam = GetNext();
                int rparam;
                if (lparam == ASM_REG) {
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_STR) {
                        instruction.cmd = BC_MOVE_R_S;
                        instruction.rparam.i = sString;
                    }
                    else if (rparam == ASM_NUM) {
                        instruction.cmd = BC_MOVE_R_N;
                        instruction.rparam.d = sNumber;
                    }
                    else if (rparam == ASM_LBL) {
                        instruction.cmd = BC_MOVE_R_L;
                        instruction.rparam.i = -sLabel;
                    }
                    else if (rparam == ASM_REG) {
                        instruction.cmd = BC_MOVE_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_MOVE);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_MOVE);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_JMP) {
                int lparam = GetNext();
                if (lparam == ASM_LBL) {
                    instruction.cmd = BC_JMP_L;
                    instruction.lparam.i = -sLabel;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_JMP);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_EVAL) {
                 int lparam = GetNext();
                 if (lparam == ASM_REG) {
                     instruction.cmd = BC_EVAL_R;
                     instruction.lparam.i = sRegister;
                 }
                 else {
                     sprintf(sError, "Error: %s, invalid lparam", ASM_EVAL);
                     sNext = ASM_ERR;
                     break;
                 }
            }
            else if (sCommand == CMD_JMPT) {
                int lparam = GetNext();
                if (lparam == ASM_LBL) {
                    instruction.cmd = BC_JMPT_L;
                    instruction.lparam.i = -sLabel;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_JMPT);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_JMPF) {
                int lparam = GetNext();
                if (lparam == ASM_LBL) {
                    instruction.cmd = BC_JMPF_L;
                    instruction.lparam.i = -sLabel;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_JMPF);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_PUSH) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_PUSH_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_PUSH);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_POP) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_POP_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_POP);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_SWAP) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_SWAP_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_SWAP);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_SPOP) {
                int lparam = GetNext();
                int rparam;
                if (lparam == ASM_REG) {
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_SPOP_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_SPOP);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_SPOP);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_OR) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_OR_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_OR);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_OR);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_AND) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_AND_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_AND);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_AND);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_NOT) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_NOT_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_NOT);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_EQL) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_EQL_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_EQL);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_EQL);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_LESS) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_LESS_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_LESS);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_LESS);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_GRE) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_GRE_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_GRE);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_GRE);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_LEQL) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_LEQL_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_LEQL);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_LEQL);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_GEQL) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_GEQL_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_GEQL);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_GEQL);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_NEQL) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_NEQL_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_NEQL);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_NEQL);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_ADD) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_ADD_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_ADD);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_ADD);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_SUB) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_SUB_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_SUB);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_SUB);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_MUL) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_MUL_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_MUL);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_MUL);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_DIV) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_DIV_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_DIV);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_DIV);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_MOD) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_MOD_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_MOD);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_MOD);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_NEG) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_NEG_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_NEG);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_CTBL) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_CTBL_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_CTBL);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_TOSTR) {
                int lparam = GetNext();
                int rparam;
                if (lparam == ASM_REG) {
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.rparam.i = sRegister;
                        /* uh ... quickfix. */
                        if (instruction.lparam.i == instruction.rparam.i) {
                            instruction.cmd = BC_STR_R;
                        }
                        else {
                            instruction.cmd = BC_STR_R_R;
                        }
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_TOSTR);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_TOSTR);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_TONUM) {
                int lparam = GetNext();
                int rparam;
                if (lparam == ASM_REG) {
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.rparam.i = sRegister;
                        if (instruction.lparam.i == instruction.rparam.i) {
                            instruction.cmd = BC_NUM_R;
                        }
                        else {
                            instruction.cmd = BC_NUM_R_R;
                        }
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_TONUM);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_TONUM);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_TOINT) {
                int lparam = GetNext();
                int rparam;
                if (lparam == ASM_REG) {
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.rparam.i = sRegister;
                        if (instruction.lparam.i == instruction.rparam.i) {
                            instruction.cmd = BC_INT_R;
                        }
                        else {
                            instruction.cmd = BC_INT_R_R;
                        }
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_TOINT);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_TOINT);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_ABS) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_ABS_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_ABS);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_MDEL) {
                int lparam = GetNext();
                if (lparam == ASM_STR) {
                    instruction.cmd = BC_MDEL_S;
                    instruction.lparam.i = sString;
                }
                else if (lparam == ASM_VAR) {
                    instruction.cmd = BC_MDEL_S;
                    instruction.lparam.i = sVariable;
                }
                else if (lparam == ASM_NUM) {
                    instruction.cmd = BC_MDEL_N;
                    instruction.lparam.i = (int)sNumber;
                }
                else if (lparam == ASM_REG) {
                    instruction.cmd = BC_MDEL_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_MDEL);
                    sNext = ASM_ERR;
                    break;
                }
                   
            }
            else if (sCommand == CMD_GC) {
                instruction.cmd = BC_GC;
            }
            else if (sCommand == CMD_LGC) {
                instruction.cmd = BC_LGC;
            }
            else if (sCommand == CMD_ULGC) {
                instruction.cmd = BC_ULGC;
            }
            else if (sCommand == CMD_CPY) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_CPY_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_CPY);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_CPY);
                    sNext = ASM_ERR;
                    break;
                 }
            }
            else if (sCommand == CMD_ECMP) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_ECMP_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_ECMP);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_ECMP);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_RTE) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_RTE_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_RTE);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_CLR) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_CLR_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_CLR);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_CALL) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_CALL_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_CALL);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_RET) {
                instruction.cmd = BC_RET;
            }
            else if (sCommand == CMD_LOCAL) {
                instruction.cmd = BC_LOCAL;
            }
            else if (sCommand == CMD_MCLR) {
                instruction.cmd = BC_MCLR;
            }
            else if (sCommand == CMD_SIZE) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_SIZE_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_SIZE);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_SIZE);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_LEN) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_LEN_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_LEN);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_LEN);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_LPTBL) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_LPTBL_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_LPTBL);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_ILOAD) {
                instruction.cmd = BC_ILOAD;
            }
            else if (sCommand == CMD_IHAS) {
                instruction.cmd = BC_IHAS;
            }
            else if (sCommand == CMD_IVAL) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_IVAL_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_IVAL);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_IKEY) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_IKEY_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_IKEY);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_IPUSH) {
                instruction.cmd = BC_IPUSH;
            }
            else if (sCommand == CMD_IPOP) {
                instruction.cmd = BC_IPOP;
            }
            else if (sCommand == CMD_ISTEP) {
                instruction.cmd = BC_ISTEP;
            }
            else if (sCommand == CMD_IDEL) {
                instruction.cmd = BC_IDEL;
            }
            else if (sCommand == CMD_COS) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_COS_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_COS);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_SIN) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_SIN_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_SIN);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_TAN) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_TAN_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_TAN);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_ACOS) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_ACOS_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_ACOS);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_ASIN) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_ASIN_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_ASIN);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_ATAN) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_ATAN_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_ATAN);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_ATAN2) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_ATAN2_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_ATAN2);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_ATAN2);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_SQR) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_SQR_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_SQR);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_POW) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_POW_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_POW);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_POW);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_FLOOR) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_FLOOR_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_FLOOR);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_CEIL) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_CEIL_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_CEIL);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_ROUND) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_ROUND_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_ROUND);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_RAD) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_RAD_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_RAD);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_DEG) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    instruction.cmd = BC_DEG_R;
                    instruction.lparam.i = sRegister;
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_DEG);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_MIN) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_MIN_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_MIN);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_MIN);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_MAX) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_MAX_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_MAX);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_MAX);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_TYPE) {
                int lparam = GetNext();
                if (lparam == ASM_REG) {
                    int rparam;
                    instruction.lparam.i = sRegister;
                    rparam = GetNext();
                    if (rparam == ASM_REG) {
                        instruction.cmd = BC_TYPE_R_R;
                        instruction.rparam.i = sRegister;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid rparam", ASM_TYPE);
                        sNext = ASM_ERR;
                        break;
                    }
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_TYPE);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else if (sCommand == CMD_SYS) {
                int lparam = GetNext();
                if (lparam == ASM_NUM) {
                    int rparam;
                    instruction.lparam.i = (int)sNumber;
                    rparam = GetNext();
                    if (rparam == ASM_NUM) {
                        instruction.cmd = BC_SYS_N_N;
                        instruction.rparam.i = (int)sNumber;
                    }
                    else {
                        sprintf(sError, "Error: %s, invalid lparam", ASM_SYS);
                        sNext = ASM_ERR;
                        break;
                    }                    
                }
                else {
                    sprintf(sError, "Error: %s, invalid lparam", ASM_SYS);
                    sNext = ASM_ERR;
                    break;
                }
            }
            else {
                strcpy(sError, "Error: Command not implemented");
                sNext = ASM_ERR;
                break;            
            }
            AddInstruction(instruction);
        }
        else if (sNext == ASM_LBL) {
            LabelEntry *lbl = HT_Get(sLabels, sLabelName, 0);
            lbl->instruction = sInstructionCount;
        }
        else if (sNext != ASM_IGNORE) {
            strcpy(sError, "Error: Expected command");
            sNext = ASM_ERR;
            break;
        }
            
        sNext = GetNext();
    }
    fclose(sFile);
    
    /* Error? */
    if (sNext == ASM_ERR) {
        CleanUp();
        return ASM_FAILURE;
    }
 
    /* Add BC_END instruction, since I always forget it. */
    Instruction end;
    end.cmd = BC_END;
    AddInstruction(end);
    
    /* Link instructions to all jumps and detect unset labels. */
    HT_ApplyKeyFunction(sLabels, LinkLabel, 0);
    if (sNext == ASM_ERR) {
        CleanUp();
        return ASM_FAILURE;
    }
    
    /* Write to binary file. */
    out = fopen(dstFilename, "wb");
    
    /* Write line number metadata. */
    fwrite(&sLineNumberCount, sizeof(int), 1, out);
    if (sLineNumberCount > 0) fwrite(sLineNumbers, sizeof(LineNumberMetadata), sLineNumberCount, out);
    
    /* Write filename metadata. */
    fwrite(&sFilenameCount, sizeof(int), 1, out);
    if (sFilenameCount > 0) {
        for (int i = 0; i < sFilenameCount; i++) {
            int len = strlen(sFilenames[i].filename);
            fwrite(&sFilenames[i].instructionIndex, sizeof(int), 1, out);
            fwrite(&len, sizeof(int), 1, out);
            fwrite(sFilenames[i].filename, sizeof(char), strlen(sFilenames[i].filename), out);
            free(sFilenames[i].filename);
        }
    }
    
    /* Copy strings from hash table to array. */
    sStringTable = (char **)malloc(sizeof(char *)*sStringCount);
    HT_ApplyKeyFunction(sStrings, StringEntryToTable, 0);
 
    /* Write strings. */
    fwrite(&sStringCount, sizeof(int), 1, out);
    for (int i = 0; i < sStringCount; i++) {
        int len = strlen(sStringTable[i]);
        fwrite(&len, sizeof(int), 1, out);
        fwrite(sStringTable[i], sizeof(char), len, out);
    }
    
    /* Write instructions. */
    fwrite(&sInstructionCount, sizeof(int), 1, out);
    if (sInstructionCount) fwrite(sInstructions, sizeof(Instruction), sInstructionCount, out);
    
    fclose(out);

    CleanUp();

    return ASM_SUCCESS;
}

/*
 * CleanUp
 * -------
 * Clean up.
 */
static void CleanUp() {
    DeleteCommands();
    
    if (sLineNumbers) {
        free(sLineNumbers);
        sLineNumbers = 0;
    }
    if (sFilenames) {
        free(sFilenames);
        sFilenames = 0;
    }
    if (sStringTable) {
        free(sStringTable);
        sStringTable = 0;
    }
    if (sStrings) {
        HT_Free(sStrings, DeleteStringEntry);
        sStrings = 0;
    }
    if (sLabels) {
        HT_Free(sLabels, DeleteLabelEntry);
        sLabels = 0;
    }
    if (sInstructions) {
        free(sInstructions);
        sInstructions = 0;
    }
}

/*
 * ASM_Error
 * ---------
 * Return error message.
 */
const char *ASM_Error() {
    return sError;
}

/*
 * StringEntryToTable
 * ------------------
 * Convert strings in hash table to table.
 */
static void StringEntryToTable(char *skey, int ikey, void *data, void *userData) {
    StringEntry *se = (StringEntry *)data;
    sStringTable[se->index] = skey;
}

/*
 * DeleteStringEntry
 * -----------------
 * Delete string entry.
 */
static void DeleteStringEntry(void *data) {
    StringEntry *se = (StringEntry *)data;
    free(se);
}

/*
 * LinkLabel
 * ---------
 * Link label indexes to instructions and check for undefined labels.
 */
static void LinkLabel(char *skey, int ikey, void *data, void *userData) {
    LabelEntry *le = (LabelEntry *)data;
    if (le->instruction < 0) {
        /* Reusing ... */
        if (sNext != ASM_ERR) {
            sNext = ASM_ERR;
            sprintf(sError, "Error: Missing label %s", skey);
        }
    }
    else {
        /* Set lparam for jumps to the instruction. */
        for (int i = 0; i < sInstructionCount; i++) {
            unsigned short cmd = sInstructions[i].cmd;
            if (cmd == BC_JMP_L || cmd == BC_JMPT_L || cmd == BC_JMPF_L ||
                cmd == BC_MSET_L) {
                /*if (sInstructions[i].rparam.i == le->index) {
                    sInstructions[i].lparam.i = le->instruction;
                }*/
                if (sInstructions[i].lparam.i < 0 && le->index == -sInstructions[i].lparam.i) {
                    sInstructions[i].lparam.i = le->instruction;
                }
            }
            else if (cmd == BC_MOVE_R_L) {
                if (sInstructions[i].rparam.i < 0 && le->index == -sInstructions[i].rparam.i) {
                   sInstructions[i].rparam.i = le->instruction;
                }
            }
        }
    }
}

/*
 * DeleteLabelEntry
 * ----------------
 * Delete label entry.
 */
void DeleteLabelEntry(void *data) {
    LabelEntry *le = (LabelEntry *)data;
    free(le);
}

/*
 * AddInstruction
 * --------------
 * Add instruction.
 */
void AddInstruction(Instruction instruction) {
    if (!sInstructions) {
        sInstructions = (Instruction *)malloc(sizeof(Instruction)*INSTRUCTIONS_GROW_SIZE);
        sInstructionsCapacity = INSTRUCTIONS_GROW_SIZE;
    }
    if (sInstructionCount >= sInstructionsCapacity) {
        sInstructionsCapacity += INSTRUCTIONS_GROW_SIZE;
        sInstructions = (Instruction *)realloc(sInstructions, sizeof(Instruction)*sInstructionsCapacity);
    }
    sInstructions[sInstructionCount++] = instruction;
}

/*
 * IsWhitespace
 * ------------
 * Return 1 if character is whitespace, else 0.
 */
int IsWhitespace(int c) {
    return c == ' ' || c == '\t' || c == '\n';
}

/*
 * AddCommand
 * ----------
 */
static void AddCommand(const char *key, Command value) {
    CommandEntry *ce = (CommandEntry *)malloc(sizeof(CommandEntry));
    ce->command = value;
    HT_Add(sCommands, key, 0, ce);
}

/*
 * CreateCommands
 * --------------
 */
static void CreateCommands() {
    sCommands = HT_Create(1);
    AddCommand(ASM_NOP, CMD_NOP);
    AddCommand(ASM_END, CMD_END);
    AddCommand(ASM_MDUMP, CMD_MDUMP);
    AddCommand(ASM_RDUMP, CMD_RDUMP);
    AddCommand(ASM_SDUMP, CMD_SDUMP);
    AddCommand(ASM_MADD, CMD_MADD);
    AddCommand(ASM_MLOAD, CMD_MLOAD);
    AddCommand(ASM_MSET, CMD_MSET);
    AddCommand(ASM_MGET, CMD_MGET);
    AddCommand(ASM_MPUSH, CMD_MPUSH);      
    AddCommand(ASM_MPOP, CMD_MPOP);
    AddCommand(ASM_MSWAP, CMD_MSWAP);
    AddCommand(ASM_MOVE, CMD_MOVE);
    AddCommand(ASM_JMP, CMD_JMP);
    AddCommand(ASM_EVAL, CMD_EVAL);
    AddCommand(ASM_JMPT, CMD_JMPT);
    AddCommand(ASM_JMPF, CMD_JMPF);
    AddCommand(ASM_PUSH, CMD_PUSH);
    AddCommand(ASM_POP, CMD_POP);
    AddCommand(ASM_SWAP, CMD_SWAP);
    AddCommand(ASM_SPOP, CMD_SPOP);
    AddCommand(ASM_OR, CMD_OR);
    AddCommand(ASM_AND, CMD_AND);
    AddCommand(ASM_NOT, CMD_NOT);
    AddCommand(ASM_EQL, CMD_EQL);
    AddCommand(ASM_LESS, CMD_LESS);
    AddCommand(ASM_GRE, CMD_GRE);
    AddCommand(ASM_LEQL, CMD_LEQL);
    AddCommand(ASM_GEQL, CMD_GEQL);
    AddCommand(ASM_NEQL, CMD_NEQL);
    AddCommand(ASM_ADD, CMD_ADD);
    AddCommand(ASM_SUB, CMD_SUB);
    AddCommand(ASM_MUL, CMD_MUL);
    AddCommand(ASM_DIV, CMD_DIV);
    AddCommand(ASM_MOD, CMD_MOD);
    AddCommand(ASM_NEG, CMD_NEG);
    AddCommand(ASM_CTBL, CMD_CTBL);
    AddCommand(ASM_TOSTR, CMD_TOSTR);
    AddCommand(ASM_TONUM, CMD_TONUM);
    AddCommand(ASM_TOINT, CMD_TOINT);
    AddCommand(ASM_ABS, CMD_ABS);
    AddCommand(ASM_MDEL, CMD_MDEL);
    AddCommand(ASM_GC, CMD_GC);
    AddCommand(ASM_LGC, CMD_LGC);
    AddCommand(ASM_ULGC, CMD_ULGC);
    AddCommand(ASM_CPY, CMD_CPY);
    AddCommand(ASM_ECMP, CMD_ECMP);
    AddCommand(ASM_RTE, CMD_RTE);
    AddCommand(ASM_CLR, CMD_CLR);
    AddCommand(ASM_CALL, CMD_CALL);
    AddCommand(ASM_RET, CMD_RET);
    AddCommand(ASM_LOCAL, CMD_LOCAL);
    AddCommand(ASM_MCLR, CMD_MCLR);
    AddCommand(ASM_LOADPM, CMD_LOADPM);
    AddCommand(ASM_SIZE, CMD_SIZE);
    AddCommand(ASM_LEN, CMD_LEN);
    AddCommand(ASM_ILOAD, CMD_ILOAD);
    AddCommand(ASM_IHAS, CMD_IHAS);
    AddCommand(ASM_IVAL, CMD_IVAL);
    AddCommand(ASM_IKEY, CMD_IKEY);
    AddCommand(ASM_IPUSH, CMD_IPUSH);
    AddCommand(ASM_IPOP, CMD_IPOP);
    AddCommand(ASM_ISTEP, CMD_ISTEP);
    AddCommand(ASM_IDEL, CMD_IDEL);
    AddCommand(ASM_COS, CMD_COS);
    AddCommand(ASM_SIN, CMD_SIN);
    AddCommand(ASM_TAN, CMD_TAN);
    AddCommand(ASM_ACOS, CMD_ACOS);
    AddCommand(ASM_ASIN, CMD_ASIN);
    AddCommand(ASM_ATAN, CMD_ATAN);
    AddCommand(ASM_ATAN2, CMD_ATAN2);
    AddCommand(ASM_SQR, CMD_SQR);
    AddCommand(ASM_POW, CMD_POW);
    AddCommand(ASM_FLOOR, CMD_FLOOR);
    AddCommand(ASM_CEIL, CMD_CEIL);
    AddCommand(ASM_ROUND, CMD_ROUND);
    AddCommand(ASM_RAD, CMD_RAD);
    AddCommand(ASM_DEG, CMD_DEG);
    AddCommand(ASM_MIN, CMD_MIN);
    AddCommand(ASM_MAX, CMD_MAX);
    AddCommand(ASM_TYPE, CMD_TYPE);
    AddCommand(ASM_MLOADS, CMD_MLOADS);
    AddCommand(ASM_LPTBL, CMD_LPTBL);
    AddCommand(ASM_SYS, CMD_SYS);
}

/* Used when searching for a command name. */
static Command sCommandSearch;
static char sFoundCommand[64];

/*
 * SearchCommand
 * -------------
 */
static void SearchCommand(char *skey, int ikey, void *data, void *userData) {
    CommandEntry *ce = (CommandEntry *)data;
    if (ce->command == sCommandSearch) {
        strcpy(sFoundCommand, skey);
    }
}

/*
 * GetCommandString
 * ----------------
 */
static const char *GetCommandString(Command command) {
    if (!sCommands) return "";
    
    sCommandSearch = command;
    strcpy(sFoundCommand, "");
    HT_ApplyKeyFunction(sCommands, SearchCommand, 0);
    return sFoundCommand;
}

/*
 * DeleteCommandEntry
 * ------------------
 */
static void DeleteCommandEntry(void *data) {
    CommandEntry *ce = (CommandEntry *)data;
    free(ce);
}

/*
 * DeleteCommands
 * --------------
 */
static void DeleteCommands() {
    if (sCommands) {
        HT_Free(sCommands, DeleteCommandEntry);
        sCommand = 0;
    }
}

/*
 * GetCommand
 * ----------
 * Return command id for string cmd or 0 if s is not a command.
 */
unsigned char GetCommand(const char *cmd) {
    CommandEntry *ce = HT_Get(sCommands, cmd, 0);
    if (ce) return ce->command;
    else return 0;
}

/*
 * GetNext
 * -------
 * Load next token and return its type.
 */
int GetNext() {
    int c = fgetc(sFile);
    while (IsWhitespace(c)) c = fgetc(sFile);
    
    /* Metadata. */
    if (c == '/') {
        char tag[8];
        int i = 0;
        c = fgetc(sFile);
        while (!(c == ':' || i > 6)) {
            tag[i++] = c;
            c = fgetc(sFile);
        }
        tag[i] = '\0';
        if (strcmp(tag, "line") == 0) {
            char line[9];
            int lineNumber;
            i = 0;
            c = fgetc(sFile);
            while (!(c == '\n' || c == EOF || i > 7)) {
                line[i++] = c;
                c = fgetc(sFile);
            }
            if (i > 7) {
                sprintf(sError, "%s, bad metadata", tag);
                return ASM_ERR;
            }
            else {
                line[i] = '\0';
                lineNumber = atoi(line);
                AddLineNumberMetadata(lineNumber);
            }
            return ASM_IGNORE;
        }
        else if (strcmp(tag, "file") == 0) {
            char line[512];
            i = 0;
            c = fgetc(sFile);
            while (!(c == '\n' || c == EOF)) {
                line[i++] = c;
                c = fgetc(sFile);
            }
            line[i] = '\0';
            AddFilenameMetadata(line);
            return ASM_IGNORE;
        }
        else {
            sprintf(sError, "Invalid metadata tag, %s", tag);
            return ASM_ERR;
        }
        
    }
    /* Register. */
    if (c == '@') {
        c = fgetc(sFile);
        if (isdigit(c)) {
            sRegister = c - 48;
            return ASM_REG;
        }
        else {
            strcpy(sError, "Error: Invalid register");
            return ASM_ERR;
        }
    }
    /* Table entry, variable. */
    else if (c == '.') {
        char varName[ASM_VAR_MAX_CHARS];
        char *s = varName;
        c = fgetc(sFile);        
        while (isalpha(c) || isdigit(c) || c == '_') {
            *s = (char)c;
            s++;
            c = fgetc(sFile);
        }
        ungetc(c, sFile);
        if (s == varName) {
            strcpy(sError, "Error: Invalid variable name");
            return ASM_ERR;
        }
        else {
            *s = '\0';
            sVariable = AddString(varName);
            return ASM_VAR;
        }
    }
    /* Number. */
    else if (isdigit(c)) {
        char num[64];
        char *s = num;
        int decimal = 0;
        while (isdigit(c) || (c == '.' && decimal++ == 0)) {
            *s = c;
            s++;
            c = fgetc(sFile);
        }
        ungetc(c, sFile);
        *s = '\0';
        sNumber = atof(num);
        
        return ASM_NUM;
    }
    /* String. */
    else if (c == '\"') {
        char txt[ASM_STRING_MAX_CHARS];
        char *s = txt;
        c = fgetc(sFile);
        while (c != '\"') {
            *s = (char)c;
            s++;
            c = fgetc(sFile);
        }
        *s = '\0';
        sString = AddString(txt);
        return ASM_STR;
    }
    /* Command or label. */
    else if (isalpha(c) || c == '_') {
        char cmd[64];
        char *s = cmd;
        while (isalpha(c) || isdigit(c) || c == '_') {
            *s = c;
            s++;
            c = fgetc(sFile);
        }
        *s = '\0';
        /* Label. */
        if (c == ':') {
            sLabel = AddLabel(cmd);
            strcpy(sLabelName, cmd);
            return ASM_LBL;
        }
        /* Command. */
        else {
            ungetc(c, sFile);
            c = GetCommand(cmd);
            if (c > 0) {
                sCommand = (unsigned char)c;
                return ASM_CMD;
            }
            else {
                sprintf(sError, "Error: %s, unknown command", cmd);
                return ASM_ERR;
            }
        }
    }
    else if (c == EOF) {
        return ASM_EOF;
    }
    else {
        sprintf(sError, "Error: Unexpected character %c", c);
        return ASM_ERR;
    }
}

/*
 * AddString
 * ---------
 * Add string str to hash table if it's not already present and return its
 * index.
 */
int AddString(const char *str) {
    StringEntry *entry = (StringEntry *)HT_Get(sStrings, str, 0);
    if (entry) return entry->index;
    
    entry = (StringEntry *)malloc(sizeof(StringEntry));
    entry->index = sStringCount++;
    HT_Add(sStrings, str, 0, entry);
    
    return entry->index;
}

/* 
 * AddLabel
 * --------
 * Add label lbl to hash table if it's not already present and return its
 * index.
 */
int AddLabel(const char *lbl) {
    LabelEntry *entry = (LabelEntry *)HT_Get(sLabels, lbl, 0);
    if (entry) return entry->index;
    
    entry = (LabelEntry *)malloc(sizeof(LabelEntry));
    entry->index = sLabelId++;
    entry->instruction = -1;
    HT_Add(sLabels, lbl, 0, entry);
    
    return entry->index;
}

void AddLineNumberMetadata(int lineNumber) {
    LineNumberMetadata lnmd;
    lnmd.instructionIndex = sInstructionCount;
    lnmd.lineNumber = lineNumber;
    if (!sLineNumbers) {
        sLineNumbers = (LineNumberMetadata *)malloc(sizeof(LineNumberMetadata)*LINE_NUMBERS_GROW_SIZE);
        sLineNumbersCapacity = LINE_NUMBERS_GROW_SIZE;
    }
    if (sLineNumberCount >= sLineNumbersCapacity) {
        sLineNumbersCapacity += LINE_NUMBERS_GROW_SIZE;
        sLineNumbers = (LineNumberMetadata *)realloc(sLineNumbers, sizeof(LineNumberMetadata)*sLineNumbersCapacity);
    }
    sLineNumbers[sLineNumberCount++] = lnmd;
}

void AddFilenameMetadata(const char *filename) {
    FilenameMetadata fnmd;
    fnmd.instructionIndex = sInstructionCount;
    fnmd.filename = strdup(filename);
    if (!sFilenames) {
        sFilenames = (FilenameMetadata *)malloc(sizeof(FilenameMetadata)*FILENAMES_GROW_SIZE);
        sFilenamesCapacity = FILENAMES_GROW_SIZE;
    }
    if (sFilenameCount >= sFilenamesCapacity) {
        sFilenamesCapacity += FILENAMES_GROW_SIZE;
        sFilenames = (FilenameMetadata *)realloc(sFilenames, sizeof(FilenameMetadata)*sFilenamesCapacity);
    }
    sFilenames[sFilenameCount++] = fnmd;
}
