/*
 * asm.h
 * -----
 * N7 assembler compiler.
 *
 * By: Marcus 2021
 */
 
#ifndef __N7_ASM_H__
#define __N7_ASM_H__
  
/* ASM_Compile return codes. */
#define ASM_SUCCESS 0
#define ASM_FAILURE 1
 
#define ASM_VAR_MAX_CHARS 64
#define ASM_STRING_MAX_CHARS 512
 
/* Instructions as strings. */
#define ASM_NOP     "NOP"
#define ASM_END     "END"
#define ASM_MDUMP   "MDUMP"
#define ASM_RDUMP   "RDUMP"
#define ASM_SDUMP   "SDUMP"
#define ASM_MADD    "MADD"
#define ASM_MLOAD   "MLOAD"
#define ASM_MLOADS  "MLOADS"
#define ASM_MSET    "MSET"
#define ASM_LPTBL   "LPTBL"
#define ASM_MGET    "MGET"
#define ASM_MPUSH   "MPUSH"      
#define ASM_MPOP    "MPOP"
#define ASM_MSWAP   "MSWAP"
#define ASM_MOVE    "MOVE"
#define ASM_JMP     "JMP"
#define ASM_EVAL    "EVAL"
#define ASM_JMPT    "JMPT"
#define ASM_JMPF    "JMPF"
#define ASM_PUSH    "PUSH"
#define ASM_POP     "POP"
#define ASM_SWAP    "SWAP"
#define ASM_SPOP    "SPOP"
#define ASM_OR      "OR"
#define ASM_AND     "AND"
#define ASM_NOT     "NOT"
#define ASM_EQL     "EQL"
#define ASM_LESS    "LESS"
#define ASM_GRE     "GRE"
#define ASM_LEQL    "LEQL"
#define ASM_GEQL    "GEQL"
#define ASM_NEQL    "NEQL"
#define ASM_ADD     "ADD"
#define ASM_SUB     "SUB"
#define ASM_MUL     "MUL"
#define ASM_DIV     "DIV"
#define ASM_MOD     "MOD"
#define ASM_NEG     "NEG"
#define ASM_CTBL    "CTBL"
#define ASM_TOSTR   "STR"
#define ASM_TONUM   "NUM"
#define ASM_TOINT   "INT"
#define ASM_ABS     "ABS"
#define ASM_MDEL    "MDEL"
#define ASM_GC      "GC"
#define ASM_LGC     "LGC"
#define ASM_ULGC    "ULGC"
#define ASM_CPY     "CPY"
#define ASM_ECMP    "ECMP"
#define ASM_RTE     "RTE"
#define ASM_CLR     "CLR"
#define ASM_CALL    "CALL"
#define ASM_RET     "RET"
#define ASM_LOCAL   "LOCAL"
#define ASM_MCLR    "MCLR"
#define ASM_LOADPM  "LOADPM"
#define ASM_SIZE    "SIZE"
#define ASM_LEN     "LEN"
#define ASM_ILOAD   "ILOAD"
#define ASM_IHAS    "IHAS"
#define ASM_IVAL    "IVAL"
#define ASM_IKEY    "IKEY"
#define ASM_IPUSH   "IPUSH"
#define ASM_IPOP    "IPOP"
#define ASM_ISTEP   "ISTEP"
#define ASM_IDEL    "IDEL"
#define ASM_COS     "COS"
#define ASM_SIN     "SIN"
#define ASM_TAN     "TAN"
#define ASM_ACOS    "ACOS"
#define ASM_ASIN    "ASIN"
#define ASM_ATAN    "ATAN"
#define ASM_ATAN2   "ATAN2"
#define ASM_SQR     "SQR"
#define ASM_POW     "POW"
#define ASM_FLOOR   "FLOOR"
#define ASM_CEIL    "CEIL"
#define ASM_ROUND   "ROUND"
#define ASM_RAD     "RAD"
#define ASM_DEG     "DEG"
#define ASM_MIN     "MIN"
#define ASM_MAX     "MAX"

#define ASM_TYPE    "TYPE"
#define ASM_SYS     "SYS"

int ASM_Compile(const char *srcFilename, const char *dstFilename);
const char *ASM_Error();
 
#endif